###################################################################################################
#
#      1D age-based P-spline model.  Used for graduating period mortality tables and extrapolating to higher ages.
#
#      Updates available from http://www.longevitas.co.uk/graduate
#
#      Copyright (c) 2011-2012 Longevitas Ltd.  Written by Stephen J. Richards and Iain D. Currie.
#
#      Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#      documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#      the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
#      and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#      1. The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#      2. THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
#      TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#      AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#      TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#      If you do not agree to the licence terms above, do not use this software,
#
#
#      Modification history:
#
#   Date	Init.	Comment
#
# 2011.07.22	SJR	Adapted from source code taken from Projections Toolkit.
# 2011.07.28	SJR	Restructured into (i) parameters, (ii) functions and (iii) main program.
# 2011.07.30	SJR	Specified URL for updates as http://www.longevitas.co.uk/graduate
# 2011.08.11	SJR	Commented out line optimising lambda and added value to plot.
# 2011.08.12	SJR	Added survival curves to output, handling of age ranges which are not a multiple of the knot spacing.
# 2011.09.15	SJR	Removal of "ncol=1" argument to as.matrix() to ensure compatibility with old versions of R.
# 2012.11.27	SJR	Improving documentation in comments.
#
###################################################################################################

###################################################################################################
#
#      Please see the accompanying file README.txt for instructions on how to use this script.
#      This is free software and neither Longevitas Ltd, Stephen Richards nor Iain Currie can
#      supply support in installing or running R.
#
###################################################################################################

###################################################################################################
#
#	Note to licensees of the Projections Toolkit: this script is a cut-down version of
#	functionality available within the Projections Toolkit.  See the chapter on 1D models
#	in the Projections Toolkit Technical Guide.
#
###################################################################################################

###################################################################################################
#
#      Parameter area - feel free to modify these values.
#
#      External filename with mortality data.
#      Note that for R you need double backslashes to separate names in the directory hierarchy.
#      Do not include data for very young ages, and certainly not for age 0.
gsInputFilename <- "C:\\Longevitas\\Graduate\\EW males in 2009.csv"

# Name of file where you want graduated rates to go
gsOutputFilename <- "C:\\Longevitas\\Graduate\\Graduated rates.csv"

# Set the upper age limit to which you want mortality rates extrapolated.
guiExtendedAge <- 120

# Knot space between splines: any positive number, preferably 5 or greater.
gdAge.dx <- 10.0

# Spline degree: any positive number, but ideally left at 3 for cubic splines.
guiBdeg <- 3

# Maximum number of iterations: any positive number.
guiMaxIterations <- 20

# Tolerance for convergence.
gdEPS <- 0.00001

# Value of lambda for smoothing: any positive number, but recommended to be at least 300.
# Note that the smaller the knot spacing, the larger the value of lambda will need to be.
gdLambda <- 1000.0

# Default value of lambda for initial fit.
gdDefaultLambda <- 1.0


###################################################################################################
#
#     Function definitions.
#
###################################################################################################
#
#     Update function to perform a single iteration of penalized scoring algorithm in one dimension
#
#     Arguments:
#
#     X.   = regression matrix in age
#     V.   = vector, data present/missing
#     OFF. = offset vector
#     D.   = vector of observed deaths
#     A.   = a matrix, ncol = 1, current estimate of regression
#            coefficients
#     P.   = penalty matrix
#
Update1D <- function(X., V., OFF., D., A., P.)
{
  Ita <- as.matrix(OFF. + X. %*% A.)
  Mu <- exp(Ita)
  Wt <- as.vector(Mu * V.)
  Bt.W.B <- t(X.) %*% (Wt * X.)
  Rhs <- Bt.W.B %*% A. + t(X.) %*% (V. * (D. - Mu))
  a <- solve(Bt.W.B + P., Rhs)
  matrix(a, ncol = 1)
}

#############################################################################
#
#      Estimate regression coefficients.  This calls the update function until convergence.
#      The user can set convergence criterion and maximum number of iterations.
#
#      Arguments
#
#      As Update1D().  Also:
#
#      A.INIT. = initial estimate of a
#      TOL. = tolerance,
#      MAX.ITER. = maximum number of iterations before aborting
#
Estimate1D <- function(X., V., OFF., D., A.INIT., P., TOL., MAX.ITER.)
{
  a <- A.INIT.
  a.old <- 10
  iter <- 0
  tol = 1
  while(tol > TOL. && iter < MAX.ITER.)
  {
    iter <- iter + 1
    a <- Update1D(X., V., OFF., D., a, P.)
    tol <- max(abs(a - a.old)/abs(a))
    a.old <- a
  }
  if (iter >= MAX.ITER.)
  {
    cat(paste("Failed to converge in Estimate1D() after ", iter, " iterations.  Convergence tolerance achieved is ", tol, " versus target of ", TOL., "."))
    stop()
  }

  matrix(a, ncol = 1)
}

###################################################################################################
#
#       Calculates and returns the following:
#
#          Calculate$a   = estimates of regression coefficients
#          Calculate$Hazard = the hazard function = Ba
#          Calculate$Tr  = trace of hat matrix
#          Calculate$Dev = deviance
#          Calculate$Aic = AIC (Akaike's Information Criterion)
#          Calculate$Bic = BIC (Bayesian Information Criterion)
#          Calculate$BtWBplusP = B'WB + P (needed for SE's)
#
#       Arguments:
#
#       A.INIT., X., V., OFF., D., TOL., MAX.ITER. as above.
#       Also:
#       P. = penalty matrix for age
#       LAMBDA. = smoothing parameter for age
#
#       Note: line 1 of the function makes sure that the
#       lambda value does not go negative.
#
Calculate1D <- function(X., V., OFF., D., A.INIT., P., LAMBDA., TOL., MAX.ITER.)
{
  lambda <- abs(LAMBDA.)
  P <- lambda * P.
  a <- Estimate1D(X., V., OFF., D., A.INIT., P, TOL., MAX.ITER.)
  Ita <- c(OFF. + X. %*% a)
  Mu <- exp(Ita)
  Wt <- Mu * V.
  Bt.W.B <- t(X.) %*% (Wt * X.)
  BtWBplusP <- Bt.W.B + P
  Rhs <- Bt.W.B %*% a + t(X.) %*% (V. * (D. - Mu))
  a <- solve(BtWBplusP, Rhs)
  Tr <- sum(diag(solve(BtWBplusP, Bt.W.B)))
  y.init <- D.
  y.init[D.==0] <- 10^(-4)
  Dev <- 2*sum( V. * (D.* log(y.init/Mu) - (D. - Mu) ) )
  Aic <- Dev + 2 * Tr
  Bic <- Dev + log(sum(V.)) * Tr
  return(list (a=a, Ita=Ita, Tr=Tr, Dev=Dev, Aic=Aic, Bic=Bic, BtWBplusP=BtWBplusP))
}

###################################################################################################
#
# Calculate and return the deviance residual
#
# For details see http://www.longevitas.co.uk/site/informationmatrix/residualconcerns.html
#
getDevianceResidual <- function(Deaths, dExpected)
{
  dResidual = 0.0

  if (Deaths < 0.0)
    stop("Negative death count not permitted!")

  if (dExpected < 0.0)
    stop("Negative expected value not permitted!")
  
  if (dExpected > 0.0)
  {
    DlogD = 0.0
    if (Deaths > 0)
      DlogD = Deaths * log(Deaths)
      
    dResidual = sign(Deaths-dExpected) * sqrt(2.0*(DlogD - Deaths*log(dExpected) - (Deaths-dExpected)))
  }

  return (dResidual)
}

###################################################################################################
#
# Main programming area - modification not recommended!
#
# Load spline library
library(splines)

# Create vector of ages.  Ages always begin at zero, even if data is not available.
gvuiAges <- seq(0, guiExtendedAge)

# Create global B-spline matrix for age.
gvdKnots <- seq(0-guiBdeg*gdAge.dx, guiExtendedAge+guiBdeg*gdAge.dx, by=gdAge.dx)
gB.a <- spline.des(gvdKnots, gvuiAges, guiBdeg+1, 0*gvuiAges, outer.ok=TRUE)$design
gc.a <- ncol(gB.a)

# Define penalty matrix - second-order penalty is hard-coded for linear extrapolation on a log scale
gP <- crossprod(diff(diff(diag(gc.a))))

# Create vectors spanning entire age range, including extrapolated ages
gvdDth <- gvuiAges * 0
gvdExp <- gvuiAges * 0.0

# Read in the data and calculate the minimum and maximum ages at which there is any data
d <- read.csv(file=gsInputFilename, header=TRUE)
for (i in 1:length(d$Age))
{
  x = d$Age[i]   
  gvdDth[x+1] = d$Deaths[i]
  gvdExp[x+1] = d$CentralExposure[i]
}
guiMinActualAge <- d$Age[1]
guiMaxActualAge <- d$Age[length(d$Age)]

# Define offset and weights vector to exclude zero-valued exposures
gvOff <- ifelse(gvdExp>0.0,log(gvdExp),0.0)
gvV <- ifelse(gvdExp>0.0,1,0)

# Find initial values
A.init <- rep(1, gc.a)
Ans <- Calculate1D(gB.a, gvV, gvOff, gvdDth, A.init, gP, gdDefaultLambda, gdEPS, guiMaxIterations)
A.init <- c(Ans$a)

# Define optimizing function
optBIC <- function(X) Calculate1D(gB.a, gvV, gvOff, gvdDth, A.init, gP, X, gdEPS, guiMaxIterations)$Bic

# We optimise values for Lambda, then calculate the fitted values for log(mu)
#gdLambda <- optim(gdDefaultLambda, optBIC, NULL, method = "L-BFGS-B", lower=0)$par
Ans <- Calculate1D(gB.a, gvV, gvOff, gvdDth, A.init, gP, gdLambda, gdEPS, guiMaxIterations)

# The fitted values of log(force of mortality) are contained in the vector gvdLogMu
gvdLogMu <- Ans$Ita - as.matrix(gvOff)

gvdCrudeMu <- ifelse(gvdExp>0.0,gvdDth / gvdExp,NA)
gvdCrudeLogMu <- ifelse(gvdCrudeMu>0,log(gvdCrudeMu),NA)
gvdFittedMu <- exp(gvdLogMu)
gvdFittedQ <- 1.0 - exp(-gvdFittedMu)

# Build vector of deviance residuals
# Remember that R indexes from 1 upwards, whereas we are counting ages from zero upwards.
gvdDevianceResiduals <- rep(NA, length(gvdCrudeLogMu))
for (i in guiMinActualAge:guiMaxActualAge)
  gvdDevianceResiduals[i+1] = getDevianceResidual(gvdDth[i+1], gvdExp[i+1]*gvdFittedMu[i+1])

# Now build actual survival curve
gvdActualSurvivalCurve <- c(rep(NA, guiMinActualAge), 1.0, exp(-cumsum(gvdCrudeMu[(guiMinActualAge+1):(guiMaxActualAge+1)])), rep(NA, guiExtendedAge-guiMaxActualAge))

# And the fitted survival curve
gvdFittedSurvivalCurve <- c(rep(NA, guiMinActualAge), 1.0, exp(-cumsum(gvdFittedMu[(guiMinActualAge+1):(guiExtendedAge+1)])))

# And now the complete life expectancy
gvdFittedCompleteLifeExpectancy <- c()
for (i in 0:guiExtendedAge)
  gvdFittedCompleteLifeExpectancy[i+1] <- 0.5+sum(cumprod(1.0-gvdFittedQ[(i+1):guiExtendedAge]))

# Now output results to file
write("Age,Deaths,CentralExposure,CrudeLogMu,DevianceResidual,ActualSurvivalCurve,GraduatedSurvivalCurve,GraduatedLogMu,GraduatedMu,GraduatedCompleteLifeExpectancy,GraduatedQ", file=gsOutputFilename, append=FALSE)
for (i in 1:(guiExtendedAge+1))
  write(paste(sep=",", gvuiAges[i], gvdDth[i], gvdExp[i], gvdCrudeLogMu[i], gvdDevianceResiduals[i], gvdActualSurvivalCurve[i], gvdFittedSurvivalCurve[i], gvdLogMu[i], gvdFittedMu[i], gvdFittedCompleteLifeExpectancy[i], gvdFittedQ[i]), file=gsOutputFilename, append=TRUE)

layout(matrix(c(1, 1, 2, 3), 2, 2, byrow=TRUE))
par(mar=c(4.5, 4.5, 0, 0), col="darkblue", col.main="darkblue", col.lab="darkblue", col.axis="darkblue", cex=1.2)
plot(x=gvuiAges, y=gvdLogMu, type="n", frame=FALSE, axes=FALSE, xlab="Age", ylab="log(mortality hazard)", xlim=range(guiMinActualAge, guiExtendedAge))
axis(1, font=2, las=1)
axis(2, font=2, las=1)
points(x=gvuiAges, y=gvdCrudeLogMu, pch=20, col="red")
lines(x=gvuiAges, y=gvdLogMu, lty=1, lwd=1, col="blue")
legend("bottomright", lwd=c(NA, NA, 1, 1), lty=c(NA, NA, NA, 1), pch=c(NA, NA, 20, NA), col=c("black", "black", "red", "blue"), legend=c(paste(sep="", "lambda=", round(gdLambda, 0)), paste(sep="", "Knot spacing=", gdAge.dx, " years"), "Crude rate", "Graduated rate"), bty="n")
legend("topleft", legend=c("Source code available at http://www.longevitas.co.uk/graduate"), bty="n", cex=0.9)

gvuiSurvivalAges <- c(gvuiAges, guiExtendedAge+1)
plot(x=gvuiSurvivalAges, y=gvdFittedSurvivalCurve, type="n", frame=FALSE, axes=FALSE, xlab="Age", ylab="Survival curve", xlim=range(guiMinActualAge, guiExtendedAge))
axis(1, font=2, las=1)
axis(2, font=2, las=1)
points(x=gvuiSurvivalAges, y=gvdActualSurvivalCurve, pch=20, col="red")
lines(x=gvuiSurvivalAges, y=gvdFittedSurvivalCurve, lty=1, col="blue")
legend("bottomleft", lwd=c(1, 1), lty=c(NA, 1), pch=c(20, NA), col=c("red", "blue"), legend=c("Crude rate", "Graduated rate"), bty="n", cex=0.9)

plot(x=gvuiAges, y=gvdDevianceResiduals, type="n", frame=FALSE, axes=FALSE, xlab="Age", ylab="Deviance residual", xlim=range(guiMinActualAge, guiMaxActualAge))
axis(1, font=2, las=1)
axis(2, font=2, las=1)
points(x=gvuiAges, y=gvdDevianceResiduals, pch=20, col="red")


###################################################################################################
#
#      Copyright (c) 2011-2012 Longevitas Ltd.  Written by Stephen J. Richards and Iain D. Currie.
#
#      Please read the licence terms at start of this file.
#
###################################################################################################
